#include <stdio.h>
#include <unistd.h>
#include <strings.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <gnome.h>
#include <pthread.h>
#include <gtk/gtk.h>

#define	COMMAND_SIZE	30

char			start_time[50];
char			old_start_time[50];
char			stop_time[50];
char			old_stop_time[50];

static FILE		*timefp;

extern int		nic;
extern gchar*		ifname;
extern int		port;
extern int		serverUp;
extern pthread_mutex_t  mutex;
extern int              pipefd[];

extern void*		cisco_thread(void *arg);
extern void*		prism2_thread(void *arg);
	     

static pthread_t	sniffer_thread_handle;

struct s{
	char	arg1[25];
	char	arg2[25];
	char	arg3[25];
};

static struct s arg_struct;	// The address of this is passed to the new thread.
				// Should it be non-static ?

/* 
 * These are the response strings that the server sends.
 * They terminate with a '\n'.
 */
static char	STOP_SNIFFING	[20];
static char 	COMMAND_FAILED	[20];
static char 	COMMAND_ACCEPTED[20];
static char	mesg		[300];

static void	update_status(void);
static int 	sendBytes(int fd, char *ptr, int nbytes);
static void 	sendFile(int sockfd, char *logfile);
static int	killSniffer(int *childpid);
static int	sendResponse(int fd, char *ptr, int nbytes);
static int	readCommand(int fd, char *ptr, int maxlen);
static int	handle_connection(int sockfd);
static void 	cleanup_WLANWatch_server(int sockfd);

void*		on_up(void *);
int		stop_sniffing;

static void update_status()
{
	pthread_mutex_lock(&mutex);
	write(pipefd[1], mesg, strlen(mesg)+1);	
	pthread_mutex_unlock(&mutex);
}

static int sendBytes(int fd, char *ptr, int nbytes)
{
	int	nleft, nwritten;

	nleft = nbytes;
	while(nleft > 0){
		nwritten = write(fd, ptr, nleft);
		if(nwritten <= 0)
			return -1;		/* error */
		nleft -= nwritten;
		ptr   += nwritten;
	}
	return (nbytes - nleft);
}

static void sendFile(int sockfd, char *logfile)
{
	char		buf[200];
	int		fd,n;
	struct stat	s;

	if( (fd=open(logfile, O_RDONLY)) == -1){
		strcpy(mesg, "sendFile failed to open logfile...\n");
		update_status();
		sendResponse(sockfd, COMMAND_FAILED, strlen(COMMAND_FAILED));
		return;
	}

	if(fstat(fd, &s) == -1){
		strcpy(mesg, "sendFile: fstat failed...\n");
		update_status();
		sendResponse(sockfd, COMMAND_FAILED, strlen(COMMAND_FAILED));
		close(fd);
		return;
	}
	sendResponse(sockfd, COMMAND_ACCEPTED, strlen(COMMAND_ACCEPTED));

	strcpy(mesg, "Exporting RxLog...\n");
	update_status();

	/* First push the file size in bytes... */
	sprintf(buf, "%d\n", (int)s.st_size);
	if(sendResponse(sockfd, buf, strlen(buf)) != strlen(buf)){
		strcpy(mesg, "Error sending size of file...\n");
		update_status();
		close(fd);
		return;
	}

	sprintf(mesg, "File size (in bytes) : %d\n", (int)s.st_size);
	update_status();

	/* Then the contents of the file */	
	while( (n = read(fd, buf, sizeof(buf))) > 0){
		if(sendBytes(sockfd, buf, n) != n){
			strcpy(mesg, "Error sending file contents...\n");
			update_status();
			close(fd);
			return;
		}
	}
	close(fd);
	strcpy(mesg, "Done.\n");
	update_status();
}

static int killSniffer(int *childpid)
{
	if(*childpid){ /* Only if the sniffer is running */
		stop_sniffing = 1;
		
		/* wait for the child to get terminated */
		if(pthread_join(sniffer_thread_handle, NULL)){
			return -1;
		}
		
		*childpid = 0;
	}
	return 0;
}

static int sendResponse(int fd, char *ptr, int nbytes)
{
	int	nleft, nwritten;

	nleft = nbytes;
	while(nleft > 0){
		nwritten = write(fd, ptr, nleft);
		if(nwritten <= 0)
			return -1;		/* error */
		nleft -= nwritten;
		ptr   += nwritten;
	}
	return (nbytes - nleft);
}

static int readCommand(int fd, char *ptr, int maxlen)
{
	int	n, rc;
	char	c;

	for(n=1; n<maxlen; n++){
		if((rc = read(fd, &c, 1)) == 1){
			if(c == '\n')
				break;
			*ptr++ = c;
		}else
			if(rc == 0){
				if(n == 1)
					return 0;	/* No data read, connection closed */
				else
					break;		/* Some data was read */
			}else
				return -1;		/* Error */
	}
	*ptr = 0;
	return n;
}

static int handle_connection(int sockfd)
{
	int		n;
	int		childpid=0;
	char		sniff_channel[3];
	char		sniff_MAC[20];
	char		buf[COMMAND_SIZE];
	char		logfile[25];
	
	/* Nullify sniff_MAC & sniff_channel */	
	sniff_channel[0] = 0;
	sniff_MAC[0] = 0;

	strcpy(start_time, 	"initializing");
	strcpy(old_start_time, 	"initializing");
	strcpy(stop_time, 	"initializing");
	strcpy(old_stop_time, 	"initializing");

	for( ; ; ){
		n = readCommand(sockfd, buf, COMMAND_SIZE);
		if(n < 0){
			strcpy(mesg, "readCommand error...\n");
			update_status();
			return -1;
		}

		if(n == 0){	/* connection closed by the client */
			return (killSniffer(&childpid));
		}

		sprintf(mesg, "Received <-- %s\n", buf);	
		update_status();

		if(!strcmp(buf, "EXIT")){
			if( (killSniffer(&childpid)) == -1){
				sendResponse(sockfd, COMMAND_ACCEPTED, strlen(COMMAND_ACCEPTED));
				return -1;
			}
			sendResponse(sockfd, COMMAND_ACCEPTED, strlen(COMMAND_ACCEPTED));
			return 0;
		}

		if(!strcmp(buf, "START_SNIFFING")){
			/* START_SNIFFING has two parameters	*/
			/* Channel Number			*/
			n = readCommand(sockfd, buf, COMMAND_SIZE);
			if(n < 0){
				strcpy(mesg, "readCommand error...\n");
				update_status();
				return -1;
			}
			if(n == 0){	/* connection closed by the client */
				return (killSniffer(&childpid));
			}
			sprintf(mesg, "Received <-- %s\n", buf);	
			update_status();
			strcpy(sniff_channel, buf);
			/* Sniff MAC				*/
			n = readCommand(sockfd, buf, COMMAND_SIZE);
			if(n < 0){
				strcpy(mesg, "readCommand error...\n");
				update_status();
				return -1;
			}
			if(n == 0){	/* connection closed by the client */
				return (killSniffer(&childpid));
			}
			sprintf(mesg, "Received <-- %s\n", buf);	
			update_status();
			strcpy(sniff_MAC, buf);

			if( (killSniffer(&childpid)) == -1){ /* 2 consecutive START_SNIFFINGs */
				sendResponse(sockfd, COMMAND_FAILED, strlen(COMMAND_FAILED));
				return -1;
			}

			timefp = popen("timenow", "r");
			if(fgets(start_time, sizeof(start_time), timefp) != NULL){
				;		
			}else
				sprintf(start_time, "%s", "timenow failed...");
			pclose(timefp);
//			printf("Start time = %s\n", start_time);

			/* start sniffer in the child process */
			strcpy(arg_struct.arg1, sniff_channel);
			strcpy(arg_struct.arg2, sniff_MAC);
			strcpy(arg_struct.arg3, ifname);
			if(nic == 0){
				if(pthread_create(&sniffer_thread_handle, NULL, cisco_thread, (void *)&arg_struct) != 0){
					strcpy(mesg, "Couldn't create cisco sniffing thread...\n");
					update_status();
					return -1; 
				}
			}else{
				if(pthread_create(&sniffer_thread_handle, NULL, prism2_thread, (void *)&arg_struct) != 0){
					strcpy(mesg, "Couldn't create prismII sniffing thread...\n");
					update_status();
					return -1;
				}
			}
			strcpy(mesg, "Sniffer Started...\n\n");
			update_status();
			
			childpid = 1;

			sendResponse(sockfd, COMMAND_ACCEPTED, strlen(COMMAND_ACCEPTED));
			continue;
		}

		if(!strcmp(buf, "STOP_SNIFFING")){
			timefp = popen("timenow", "r");
			if(fgets(stop_time, sizeof(stop_time), timefp) != NULL){
				;		
			}else
				sprintf(stop_time, "%s", "timenow failed...");
			pclose(timefp);
//			printf("Stop time  = %s\n\n", stop_time);
			
			if( (killSniffer(&childpid)) == -1){
				sendResponse(sockfd, COMMAND_FAILED, strlen(COMMAND_FAILED));
				return -1;
			}
			sendResponse(sockfd, COMMAND_ACCEPTED, strlen(COMMAND_ACCEPTED));
			continue;
		}

		if(!strcmp(buf, "IMPORT_RXLOG")){
			if(childpid){	/* Sniffer running */
				sendResponse(sockfd, STOP_SNIFFING, strlen(STOP_SNIFFING));
				continue;
			}
			strcpy(logfile, sniff_MAC);
			strcat(logfile, ".log");
			/* Send file */
			sendFile(sockfd, logfile);
			continue;
		}
	}
}

void cleanup_WLANWatch_server(int sockfd)
{
	strcpy(mesg, "Press DOWN to continue...\n");
	update_status();
	close(sockfd);
	pthread_exit(NULL);
}
			
void* on_up(void *arg)
{
	struct sockaddr_in 	cli_addr, serv_addr;
	int 			sockfd, newsockfd, cli_addr_len;
	sigset_t		mask;

	sigfillset(&mask);
	pthread_sigmask(SIG_BLOCK, &mask, NULL);
	
	strcpy(STOP_SNIFFING, 	"STOP_SNIFFING\n");
	strcpy(COMMAND_FAILED, 	"COMMAND_FAILED\n");
	strcpy(COMMAND_ACCEPTED,"COMMAND_ACCEPTED\n");

	if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		strcpy(mesg, "socket creation failed...\n"); 
		update_status();
		
		strcpy(mesg, "WLANWatch Server Terminating...\n");
		update_status();

		cleanup_WLANWatch_server(sockfd);
	}
  
	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family      = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port        = htons(port);
  
	if(bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
		sprintf(mesg, "WLANWatch Server : Bind on port %d failed...\n", port);
		update_status();

		strcpy(mesg, "WLANWatch Server Terminating...\n");
		update_status();

		cleanup_WLANWatch_server(sockfd);
	}
  
	if(listen(sockfd, 5) < 0) {
		strcpy(mesg, "listen failed...\n");
		update_status();
		strcpy(mesg, "WLANWatch Server Terminating...\n");
		update_status();

		cleanup_WLANWatch_server(sockfd);
	}

	/* To make sure that the accept call below doesnt block */
	fcntl(sockfd,  F_SETFL, O_NONBLOCK);
	
	sprintf(mesg, "WLANWatch Server Running On Port : %5d ...\n", port);
	update_status();

	cli_addr_len = sizeof(cli_addr);
	while(serverUp){
		newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &cli_addr_len);
		if(newsockfd < 0){
			continue;
		}
		
		stop_sniffing = 0;

		strcpy(mesg, "\n{Connection Accepted...}\n");
		update_status();

		strcpy(mesg, "----------------------------\n");
		update_status();

		if(handle_connection(newsockfd)){
			strcpy(mesg, "{ABNORMAL Connection Termination...}\n");
			update_status();

			cleanup_WLANWatch_server(sockfd);
		}else{
			strcpy(mesg, "{Connection Closed...}\n");
			update_status();

			strcpy(mesg, "-------------------------\n");
			update_status();
		}
		close(newsockfd);
	}
	close(sockfd);
	pthread_exit(NULL);
}
