#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

extern gint		nic;
extern gchar		*ifname;
extern gint		port;

extern gchar		*c_ifname;
extern gchar		*c_driver;
extern gchar		*c_ip;

extern gchar		*p_ifname;
extern gchar		*p_driver;
extern gchar		*p_ip;
/*
 * extern gchar		*p_ssid;
 */

extern int		serverUp;
extern GtkWidget	*WLANWatch_Server;
extern pthread_mutex_t	mutex;
extern int		pipefd[];

extern void* 		on_up(void *arg);

static pthread_t	on_up_thread;
static char		mesg[300];

static void 		read_ifname(void);
static void 		read_port(void);

static void 		read_c_ifname(void);
static void 		read_c_driver(void);
static void 		read_c_ip(void);

static void 		read_p_ifname(void);
static void 		read_p_driver(void);
static void 		read_p_ip(void);
/*
 * static void 		read_p_ssid(void);
 */

static void 		read_ALL(void);

static void read_ifname(void)
{
	GtkWidget	*text_entry = NULL;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "ifname");
	ifname		= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(ifname);
}

static void read_port(void)
{
	GtkWidget	*text_entry = NULL;
	gchar		*port_str;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "port");
	port_str	= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(port_str);

	port		= atoi(port_str);
}

static void read_c_ifname(void)
{
	GtkWidget	*text_entry = NULL;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "c_ifname");
	c_ifname		= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(c_ifname);
}

static void read_c_driver(void)
{
	GtkWidget	*text_entry = NULL;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "c_driver");
	c_driver		= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(c_driver);
}

static void read_c_ip(void)
{
	GtkWidget	*text_entry = NULL;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "c_ip");
	c_ip		= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(c_ip);
}

static void read_p_ifname(void)
{
	GtkWidget	*text_entry = NULL;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "p_ifname");
	p_ifname		= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(p_ifname);
}

static void read_p_driver(void)
{
	GtkWidget	*text_entry = NULL;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "p_driver");
	p_driver		= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(p_driver);
}

static void read_p_ip(void)
{
	GtkWidget	*text_entry = NULL;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "p_ip");
	p_ip		= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(p_ip);
}

/*
static void read_p_ssid(void)
{
	GtkWidget	*text_entry = NULL;

	text_entry 	= lookup_widget(GTK_WIDGET(WLANWatch_Server), "p_ssid");
	p_ssid		= gtk_editable_get_chars(GTK_EDITABLE(text_entry), 0, -1);
	
	g_strstrip(p_ssid);
}
*/

static void read_ALL(void)
{
	read_ifname();
	read_port();
	read_c_ifname();
	read_c_driver();
	read_c_ip();
	read_p_ifname();
	read_p_driver();
	read_p_ip();
/*
 * 	read_p_ssid();
 */
}

gboolean
on_WLANWatch_Server_delete_event             (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
	on_close_clicked(NULL, NULL);
}


void
on_help_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{

}

void
on_up_clicked                          (GtkButton       *button,
                                        gpointer         user_data)
{
	if(serverUp){
		strcpy(mesg, "WLANWatch-Server Already UP...\n");
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
		return;
	}

	read_ALL();

	serverUp = 1;

	if(pthread_create(&on_up_thread, NULL, on_up, NULL) != 0){
		strcpy(mesg, "Couldn't create on_up thread...\n");
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
		serverUp = 0;
		return;
	}
}

void
on_down_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
	if(!serverUp){
		strcpy(mesg, "WLANWatch Server Not Running...\n");
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
		return;
	}

	serverUp = 0;

	if(pthread_join(on_up_thread, NULL)){
		strcpy(mesg, "pthread_join failed on on_up thread...\n");
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
		return;
	}

	strcpy(mesg, "WLANWatch Server Down...\n");
	pthread_mutex_lock(&mutex);
	write(pipefd[1], mesg, strlen(mesg)+1);
	pthread_mutex_unlock(&mutex);
}


void
on_close_clicked                       (GtkButton       *button,
                                        gpointer         user_data)
{
	FILE *fp;

	fp = fopen(".WLANWatch-server.conf", "w");
	if(fp == NULL){
		gtk_main_quit();
	}else{
		read_ALL();
		
		fprintf(fp, "%d\n", nic);
		fprintf(fp, "%s\n", ifname);
		fprintf(fp, "%d\n", port);

		fprintf(fp, "%s\n", c_ifname);
		fprintf(fp, "%s\n", c_driver);
		fprintf(fp, "%s\n", c_ip);

		fprintf(fp, "%s\n", p_ifname);
		fprintf(fp, "%s\n", p_driver);
		fprintf(fp, "%s\n", p_ip);
		/*
		 * fprintf(fp, "%s\n", p_ssid);
		 */
		fclose(fp);
	}
	gtk_main_quit();
}

void
on_up_c_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
	char		command[200];
	FILE		*fp;

	read_ALL();

	sprintf(command, "modprobe %s", c_driver);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "WLANWatch-Server could not modprobe the driver module : %s\n", c_driver);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);

	sprintf(command, "ifconfig %s %s", c_ifname, c_ip);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}

	sprintf(command, "ifconfig %s mtu 2292", c_ifname);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}

	pclose(fp);
}

void
on_down_c_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
	char		command[200];
	FILE		*fp;

	read_ALL();

	sprintf(command, "ifconfig %s down", c_ifname);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);

	sprintf(command, "rmmod -r %s", c_driver);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);
}

void
on_up_p_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
	char		command[200];
	FILE		*fp;

	read_ALL();
	
	sprintf(command, "modprobe %s", p_driver);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "WLANWatch-Server could not modprobe the driver module : %s\n", p_driver);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);

	sprintf(command, "wlanctl-ng %s lnxreq_ifstate ifstate=disable", p_ifname);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);

	sprintf(command, "wlanctl-ng %s lnxreq_ifstate ifstate=enable", p_ifname);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);
/*
	sprintf(command, "wlanctl-ng %s lnxreq_autojoin ssid=\"%s\" authtype=opensystem", p_ifname, p_ssid);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);
*/
	sprintf(command, "ifconfig %s %s", p_ifname, p_ip);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}

	sprintf(command, "ifconfig %s mtu 2292", p_ifname);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}

	pclose(fp);
}

void
on_down_p_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
	char		command[200];
	FILE		*fp;

	read_ALL();

	sprintf(command, "wlanctl-ng %s lnxreq_ifstate ifstate=disable", p_ifname);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);

	sprintf(command, "ifconfig %s down", p_ifname);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);

	sprintf(command, "rmmod -r %s", p_driver);
	if( (fp=popen(command, "r")) == NULL){
		sprintf(mesg, "%s : failed...\n", command);
		pthread_mutex_lock(&mutex);
		write(pipefd[1], mesg, strlen(mesg)+1);
		pthread_mutex_unlock(&mutex);
	}
	pclose(fp);
}

void
on_close_c_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
	on_close_clicked(NULL, NULL);
}

void
on_close_p_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
	on_close_clicked(NULL, NULL);
}

