#include "sniffer.h"
#include <gnome.h>
#include <pthread.h>

extern gint             nic;
extern int		stop_sniffing;
extern pthread_mutex_t  mutex;
extern int              pipefd[];
extern char		start_time[];
extern char		old_start_time[];
extern char		stop_time[];
extern char		old_stop_time[];

static long int		SOCKET_BUFFER_SIZE = 524287;

static int		signallevel;
static int		noiselevel;
static unsigned char 	*MACH_ptr;
static unsigned char	*PrismH_ptr;

static MgmtFrame_1	*M1;
static MgmtFrame_2	*M2;
static PrismHeader	*prismheader;

static int 		SIZE_OF_FCS = 0;
static int 		FCSerr;
static char		FCSerrTag[5];

static int 		sniffer_socket;
static int		sniff_channel;

/* Packet sniffed through the socket	*/
static unsigned char 	packet			[MAX_BUFFER_SIZE];  

/* Size of the sniffed packet		*/
static int 		len;

static char 		sniff_MAC[18];
static char 		ifname[20];

static char		logfile[25];
static FILE		*log;
static char		mesg[300];

static void		update_status(void);
static void 		dump_to_file(unsigned char* data);
static void 		process_DATA_packet(void);
static void 		extract_header_info(void);
static void 		check_FCSerror();
static void 		nullify_pointers(void);
static int 		process_packet(void);
static int 		sniff_packet(unsigned char *buf, int maxlen);
static int 		set_channel(int channel, int true);
static int 		socket_init(void);
static void 		cleanup_sniffer();

struct r{
	char    arg1[25];
	char    arg2[25];
	char    arg3[25];
};

static struct r *arg_s;

void*		cisco_thread(void *arg);
void* 		prism2_thread(void *arg);

static void update_status()
{
	pthread_mutex_lock(&mutex);
	write(pipefd[1], mesg, strlen(mesg)+1);
	pthread_mutex_unlock(&mutex);
}

static void dump_to_file(unsigned char* data)
{
	unsigned int	packet_length;
	int		bytes_left;
	int 		i;

	/* I am interested in packets originating only from the sniff_MAC	*/
	if(strcmp(sniff_MAC,Res.SA))
		return;

	/* 
	 * I am interested in packets being transmitted only by the sniff_MAC.	
	 * TODS == 1 && FROMDS == 0 ensures that the packet is being sent from a 
	 * station to an Access Point. 
	 *
	 * Hence when this constraint is clubbed with the above one, you get only 
	 * all those packets being transmitted by the sniff_MAC.
	 */

	if(Res.toDS != 1 || Res.fromDS != 0)
		return;

	packet_length = (packet+len) - (data) - (SIZE_OF_FCS);
	bytes_left = packet_length;

	/* LLC is of 8 bytes */
	if(packet_length < 8)
		return;
	
	if(FCSerr)
		fprintf(log,"%s\n", FCSerrTag);
	
	/* Add the signal level for each packet */
	fprintf(log,"%X %X\n", signallevel, noiselevel);
	
	for(i=1; i<=4; i++){
		if(*data <= 0xf)	fprintf(log,"0%1x", *data);
		else			fprintf(log,"%2x", *data);
		data++;
		bytes_left--;

		if(*data <= 0xf)	fprintf(log,"0%1x", *data);
		else			fprintf(log,"%2x", *data);
		data++;
		bytes_left--;
		
		if(i==4)	fprintf(log,"\n");
		else		fprintf(log," ");	
	}

	i=0;
	while(bytes_left > 0){
		i++;
		if(*data <= 0xf)	fprintf(log,"0%1x", *data);
		else			fprintf(log,"%2x", *data);
		data++;
		bytes_left--;

		if(bytes_left){
			if(i==16){		
				fprintf(log,"\n");
				i=0;
			}
			else
				if(i%2==0)
					fprintf(log," ");
		}
	}
	fprintf(log,"\n\n");
}

static void process_DATA_packet()
{
	unsigned char*	data;

	// This is where the data starts in most of the following cases.
	data = MACH_ptr + sizeof(MgmtFrame_1);
	
	if (Res.toDS && Res.fromDS){
		// This is where the data starts in this case.
		data = MACH_ptr + sizeof(MgmtFrame_2);
		
		M2 = (MgmtFrame_2 *) MACH_ptr;
		sprintf (Res.RA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M2->Address1[0], M2->Address1[1], M2->Address1[2], 
			M2->Address1[3], M2->Address1[4], M2->Address1[5]);
		sprintf (Res.TA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M2->Address2[0], M2->Address2[1], M2->Address2[2], 
			M2->Address2[3], M2->Address2[4], M2->Address2[5]);
		sprintf (Res.DA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M2->Address3[0], M2->Address3[1], M2->Address3[2], 
			M2->Address3[3], M2->Address3[4], M2->Address3[5]);
		sprintf (Res.SA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M2->Address4[0], M2->Address4[1], M2->Address4[2], 
			M2->Address4[3], M2->Address4[4], M2->Address4[5]);
	}
	if (!Res.toDS && !Res.fromDS){	// IBSS	
		sprintf (Res.RA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address1[0], M1->Address1[1], M1->Address1[2], 
			M1->Address1[3], M1->Address1[4], M1->Address1[5]);
		sprintf (Res.SA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address2[0], M1->Address2[1], M1->Address2[2], 
			M1->Address2[3], M1->Address2[4], M1->Address2[5]);
		sprintf (Res.BSSID, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address3[0], M1->Address3[1], M1->Address3[2], 
			M1->Address3[3], M1->Address3[4], M1->Address3[5]);
		strcpy(Res.DA,Res.RA);
	}
	if (!Res.toDS && Res.fromDS){	// From the AP
		sprintf (Res.RA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address1[0], M1->Address1[1], M1->Address1[2], 
			M1->Address1[3], M1->Address1[4], M1->Address1[5]);
		sprintf (Res.BSSID, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address2[0], M1->Address2[1], M1->Address2[2], 
			M1->Address2[3], M1->Address2[4], M1->Address2[5]);
		sprintf (Res.SA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address3[0], M1->Address3[1], M1->Address3[2], 
			M1->Address3[3], M1->Address3[4], M1->Address3[5]);
		strcpy(Res.DA,Res.RA);
	}
	if (Res.toDS && !Res.fromDS){	// To the AP
		sprintf (Res.RA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address1[0], M1->Address1[1], M1->Address1[2], 
			M1->Address1[3], M1->Address1[4], M1->Address1[5]);
		sprintf (Res.SA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address2[0], M1->Address2[1], M1->Address2[2], 
			M1->Address2[3], M1->Address2[4], M1->Address2[5]);
		sprintf (Res.DA, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			M1->Address3[0], M1->Address3[1], M1->Address3[2],
			M1->Address3[3], M1->Address3[4], M1->Address3[5]);
		strcpy(Res.BSSID,Res.RA);
	}

	/* 
	 * NOTE: IF THE HOST IS RESPONSIBLE FOR ENCRYPTION/DECRYPTION THEN:
	 * IV (of 32 bits) follows this MAC header (ie., before LLC & IP header).
	 * ICV (of 32 bits) also appears before the FCS value in the data part.	
	 * checkout for "cnfWEPFlags" for more details. 
	 * As of now, we assume that the firmware handles the ENC/DEC & hence 
	 * there are no IV & ICV in the data.
	 */

	dump_to_file(data);
}

static void extract_header_info()
{
	int		field;

	memset(&Res, 0, sizeof (Res));

	field 			= 	M1->FrameControl;

	Res.order 		=	(field & (1 << 15)) ? 1 : 0;
	Res.wep 		=	(field & (1 << 14)) ? 1 : 0;	
	Res.moreData 		= 	(field & (1 << 13)) ? 1 : 0;
	Res.powerMgmt 		= 	(field & (1 << 12)) ? 1 : 0;
	Res.retry 		= 	(field & (1 << 11)) ? 1 : 0;
	Res.moreFlags 		= 	(field & (1 << 10)) ? 1 : 0;
	Res.fromDS 		=	(field & (1 <<  9)) ? 1 : 0;
	Res.toDS 		=	(field & (1 <<  8)) ? 1 : 0;
	Res.subType[3]		= 	(field & (1 <<  7)) ? 1 : 0;
	Res.subType[2]		= 	(field & (1 <<  6)) ? 1 : 0;
	Res.subType[1]		= 	(field & (1 <<  5)) ? 1 : 0;
	Res.subType[0]		= 	(field & (1 <<  4)) ? 1 : 0;
	Res.type[1] 		= 	(field & (1 <<  3)) ? 1 : 0;
	Res.type[0] 		= 	(field & (1 <<  2)) ? 1 : 0;
	Res.protocolVersion [1] = 	(field & (1 <<  1)) ? 1 : 0;
	Res.protocolVersion [0] = 	(field & (1 <<  0)) ? 1 : 0;
	
	signallevel		=	0;
	noiselevel		=	0;
	if(nic == 1){ // only for prism2 cards.
		signallevel	= 	prismheader->signal.data;
		noiselevel 	= 	prismheader->noise.data;
		
	/*	
		field		=	prismheader->signal.data;
		i		=	0;
	 	while (field / 10 != 0){
			signallevel	+= (field % 10) * (1 << (4*i)); // power (16, i);
			field  		/= 10;
			i++;
		}
		signallevel	= signallevel + field * (1 << (4*i));    // power (16, i);
	*/
	}
}

/* 
 * Check for FCSerr. 
 * For wlanng driver:
 * ==================
 * FCS is set to 
 * 	0x12121212 if the frame was received with FCSerr set
 * 	0xFFFFFFFF if the frame was received with FCSerr unset 
 * in the monitor mode.
 *
 * NOTE: the driver has been modified for this to happen.
 * The driver now sets the FCS to 0x12121212 if frame is received
 * with an error (in the monitor mode).
 *
 * For Cisco driver:
 * =================
 * Only uncorrupted packets are passed up.
 * Hence no FCSerr.
 */
static void check_FCSerror()
{
	unsigned char*	ch;
	int		i;

	if(nic == 0){
		FCSerr = 0;
		return;
	}
	ch = packet + len - SIZE_OF_FCS;
	FCSerr = 1;
	for(i=0; i<4; i++){
		if( *(ch+i) != 0x12)	
			FCSerr = 0;
	}
}

static void nullify_pointers()
{
	PrismH_ptr      = NULL;
	prismheader     = NULL;
       	MACH_ptr	= NULL;
	M1		= NULL;	
	M2 		= NULL;	
}

static int process_packet()
{
	int 		frame_type;

	nullify_pointers();

	if(nic == 1){	// Prism2 cards.
		/* FATAL: Length of packet sniffed is less than the size of the Prismheader */
		if (len < sizeof(PrismHeader)){
			return 0;
		}

		/* PrismH_ptr points to the start of prismheader.       */
		PrismH_ptr      = packet;

		/* MACH_ptr points to the start of 802.11b MAC header.	*/
		MACH_ptr        = &packet[sizeof(PrismHeader)];

		/* First the info from the card.                        */
		prismheader     = (PrismHeader *) PrismH_ptr;
	}else{		// Cisco cards.
		/* MACH_ptr points to the start of 802.11b MAC header.	*/
		MACH_ptr 	= packet;
	}
		        
	/* Then the actual MAC Frame.                           */
	M1              = (MgmtFrame_1 *) MACH_ptr;

	/* Check for FCSerr. 					*/
	check_FCSerror();
	
	/* Extract the details the MAC header carries into a structure	*/	
	extract_header_info();
	
	frame_type = 2*Res.type[0] + Res.type[1];
	// NOTE: this "frame_type" is just the reverse of what the 802.11b standards talk about.
	// The standards list the values with MSBs first. eg: b3b2
	if(frame_type == 1){
		process_DATA_packet();
	}	
	return 0;
}

/*
 * sniff_packet() gets a packet from the wireless
 * and dumps it onto memory.
 *
 * parameters:
 * buf - allocated buffer to put packet into
 * maxlen - length of buffer 
 * 
 */
static int sniff_packet(unsigned char *buf, int maxlen)
{
	fd_set 			rs;
	int 			r;
	struct timeval 		tv;

	FD_ZERO(&rs);
	FD_SET(sniffer_socket, &rs);
	tv.tv_sec = 10;
	tv.tv_usec = 0;

	r = select(sniffer_socket + 1, &rs, NULL, NULL, &tv);

	if(r < 0){
		strcpy(mesg, "Sniffer: select failed...\n");
		update_status();
		return -1;
	}
	if(r == 0){
		strcpy(mesg, "Sniffer: select Timeout...\n");
//		update_status();
//		return -1;
		return 0;
	}

	if(FD_ISSET(sniffer_socket, &rs)){
        	r = recv(sniffer_socket, buf, maxlen, 0);
	       	if(r < 0){
			strcpy(mesg, "Sniffer: recv failed\n");
			update_status();
			return -1;
		}
		return r;
	}
	return 0;                   /* Nothing to read */
}

/* set_channel () sets/unsets a particular channel into the monitor mode.
 *
 * parameters : 
 * channel - channel number to set/unset
 * true - if 1, sets
 * 	- if 0, unsets
 * nic	- 0 : Cisco
 * 	- 1 : PrismII
 */
static int set_channel(int channel, int true)
{
	int 		result;
  	FILE 		*fp;
  	unsigned char 	returned_string[1000];
  	unsigned char 	command[125];
	
	switch(nic){
		/* Cisco	*/
		case 0:
			/* As of now, the cisco drivers do not pass up the FCS */
			SIZE_OF_FCS = 0;
			if(true){
				sprintf(command, "echo 'Mode: rfmon' > /proc/driver/aironet/%s/Config", ifname);
				fp = popen(command, "r");
				pclose(fp);

				/*
				 * sprintf(command, "echo 'Mode: y' > /proc/driver/aironet/eth1/Config");
				 * fp = popen(command, "r");
				 * pclose(fp);
				 */
				sprintf(command, "echo 'Channel: %d' > /proc/driver/aironet/%s/Config", channel, ifname);
				fp = popen(command, "r");
				pclose(fp);
			}else{
				sprintf(command, "echo 'Mode: ESS' > /proc/driver/aironet/%s/Config", ifname);
				fp = popen(command, "r");
				pclose(fp);

				sprintf(command, "echo 'Channel: %d' > /proc/driver/aironet/%s/Config", channel, ifname);
				fp = popen(command, "r");
				pclose(fp);
			}
			break;

		/* PrismII	*/
		case 1:
			if(true){
				sprintf(command, "wlanctl-ng %s lnxreq_wlansniff enable=true channel=%d prismheader=true stripfcs=false", 
						ifname, channel);
				SIZE_OF_FCS = 4;
				/* 
				 * NOTE: For the wlan-ng drivers with version < 0.1.15-pre8,
				 * stripfcs = true : when set, passes the fcs from the driver to above and 
				 * the SIZE_OF_FCS is 4. else the SIZE_OF_FCS should be 0.
				 * By default, the fcs is *not* passed up to the applications.
				 *
				 */
			}else{
				sprintf(command, "wlanctl-ng %s lnxreq_wlansniff enable=false channel=%d prismheader=false", ifname, channel);
			}

			fp = popen(command, "r");
			if(fp < 0){
				strcpy(mesg, "Sniffer: Could not popen wlanctl-ng...\n");
				update_status();
				return -1;
			}
			
			result = fscanf (fp, "%s", returned_string);
			if((result < 0) || (result == 0)){ // nothing on stdout, on stderr ?
				strcpy(mesg, "Sniffer: Enabling/Disabling of Monitor mode Failed...\n");
				update_status();
				pclose(fp);
				return -1;
			}
			while(result > 0){
				result = fscanf(fp, "%s", returned_string);
				if (strstr((const char*)returned_string, (const char*)"resultcode=success") != NULL){
					break;
				}else 
					if(strstr(returned_string, "resultcode=refused") != NULL){
						strcpy(mesg, "Sniffer: Enabling/Disabling of Monitor mode Failed...\n");
						update_status();
						pclose(fp);
						return -1;
					}
			}
			pclose(fp);
			break;
	} /* switch */
	return 0;
}

/* socket_init() initialises the raw socket used for sniffing.	*/
static int socket_init()
{
	int ret_socket;
	int myset = 10;  // socket option to reuse address
	int result;

	switch(nic){
		/* Cisco	*/
		case 0:
			ret_socket = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_802_2));
			break;

		/* PrismII	*/
		case 1:
			ret_socket = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ECONET+1));
			break;
			/* 
			 * The wlanng driver sets the protocol type to "htons(ETH_P_80211_RAW)".
			 * and ETH_P_80211_RAW is defined as ETH_P_ECONET + 1
			 * i.e., ETH_P_80211_RAW = 0x0019 
			 * Now, no need to bind this to any interface I guess... :-)
			 */
	}
	if(ret_socket < 0){
		strcpy(mesg, "Sniffer: Could not create socket...\n");
		update_status();
		return -1;
	}

	/* Set socket options to specify receive timeout and enable reuseaddr.	*/
/*	if ((result=setsockopt(ret_socket, SOL_SOCKET, SO_REUSEADDR, &myset, sizeof(myset))) < 0){
		strcpy(mesg, "Sniffer: Could not set socket options...\n");
		update_status();
		return -1;
	}
*/
	if ((result=setsockopt(ret_socket, SOL_SOCKET, SO_RCVBUF, (char *)&SOCKET_BUFFER_SIZE, sizeof(long int))) < 0){
		strcpy(mesg, "Sniffer: Could not set socket options...\n");
		update_status();
		return -1;
	}
	return ret_socket;
}

/* 
 * nic	- 0 : Cisco
 * 	- 1 : PrismII
 */	
static void cleanup_sniffer()
{
	set_channel(sniff_channel, 0);
	
	while(!strcmp(stop_time, old_stop_time)){
//		write(1,"Stop time old one ", 18);
		;
	}
	fprintf(log, "%s\n", stop_time);
	strcpy(old_stop_time, stop_time);	

	funlockfile(log);
	fclose(log);
	pthread_exit((void *)NULL);
}

void* cisco_thread(void *arg)
{
	sigset_t                mask;

	sigfillset(&mask);
	pthread_sigmask(SIG_BLOCK, &mask, NULL);
			
	arg_s = (struct r*) arg;

	sniff_channel = atoi(arg_s->arg1);
	strcpy(sniff_MAC, arg_s->arg2);
	strcpy(ifname, arg_s->arg3);
	
	strcpy(logfile, sniff_MAC);
	strcat(logfile, ".log");
	log = fopen(logfile, "w+");
	flockfile(log);

	while(!strcmp(start_time, old_start_time)){
//		write(1,"Start time old one ", 19);
		;
	}
	fprintf(log, "%s\n", start_time);
	strcpy(old_start_time, start_time);	
	
	strcpy(FCSerrTag, "*");

	/* Initialize the socket for sniffing 			*/
	sniffer_socket = socket_init();
	if(sniffer_socket == -1){
		strcpy(mesg, "Terminating Sniffer...\n");
		update_status();
		cleanup_sniffer();
	}

	if(set_channel(sniff_channel, 1) == -1){
		strcpy(mesg, "Terminating Sniffer...\n");
		update_status();
		cleanup_sniffer();
	}
	
	/* Until the parent thread wants....			*/
	while(!stop_sniffing) {
		memset(packet, 0, MAX_BUFFER_SIZE);

		/* Sniff a packet through the socket	*/
		len = sniff_packet(packet, MAX_BUFFER_SIZE);
		if(len == -1){
			strcpy(mesg, "Terminating Sniffer...\n");
			update_status();
			cleanup_sniffer();
		}
		if(len == 0)
			continue;
			
		/* Process the packet sniffed		*/
		process_packet();
	}
	strcpy(mesg, "Sniffer Stopped...\n\n");
	update_status();

	cleanup_sniffer();
	return NULL;
}

void* prism2_thread(void *arg)
{
	sigset_t                mask;

	sigfillset(&mask);
	pthread_sigmask(SIG_BLOCK, &mask, NULL);
			
	arg_s = (struct r*) arg;

	sniff_channel = atoi(arg_s->arg1);
	strcpy(sniff_MAC, arg_s->arg2);
	strcpy(ifname, arg_s->arg3);
	
	strcpy(logfile, sniff_MAC);
	strcat(logfile, ".log");
	log = fopen(logfile, "w+");
	flockfile(log);

	while(!strcmp(start_time, old_start_time)){
//		write(1,"Start time old one ", 19);
		;
	}
	fprintf(log, "%s\n", start_time);
	strcpy(old_start_time, start_time);	
		
	strcpy(FCSerrTag, "*");

	/* Initialize the socket for sniffing 			*/
	sniffer_socket = socket_init();
	if(sniffer_socket == -1){
		strcpy(mesg, "Terminating Sniffer...\n");
		update_status();
		cleanup_sniffer();
	}

	if(set_channel(sniff_channel, 1) == -1){
		strcpy(mesg, "Terminating Sniffer...\n");
		update_status();
		cleanup_sniffer();
	}
	
	/* Until the parent thread wants....			*/
	while(!stop_sniffing) {
		memset(packet, 0, MAX_BUFFER_SIZE);

		/* Sniff a packet through the socket	*/
		len = sniff_packet(packet, MAX_BUFFER_SIZE);
		if(len == -1){
			strcpy(mesg, "Terminating Sniffer...\n");
			update_status();
			cleanup_sniffer();
		}
		if(len == 0)
			continue;
			
		/* Process the packet sniffed		*/
		process_packet();
	}
	strcpy(mesg, "Sniffer Stopped...\n\n");
	update_status();

	cleanup_sniffer();
	return NULL;
}

